from django.conf import settings
from django.db import models
from tenant_schemas.models import TenantMixin


class MultiTenantMixin(TenantMixin):
    """Restrict only one inheritance level."""

    tenant = models.ForeignKey(
        settings.TENANT_MODEL,
        related_name='tenants',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.schema_name

    @property
    def codename(self):
        return self.id

    def is_leading(self):
        """Check if it's main tenant."""
        return not self.tenant_id



import os

from multi_tenant.migration_executors.base import MigrationExecutor
from multi_tenant.migration_executors.standard import StandardExecutor


def get_executor(codename=None):
    codename = codename or os.environ.get('EXECUTOR', StandardExecutor.codename)

    for klass in MigrationExecutor.__subclasses__():
        if klass.codename == codename:
            return klass

    raise NotImplementedError(f'No executor with codename {codename}')



from django.conf import settings
from django.core.management import (
    call_command,
    get_commands,
    load_command_class
)
from django.core.management.base import BaseCommand, CommandError
from django.db import connection
from tenant_schemas.utils import get_public_schema_name, get_tenant_model


class BaseTenantCommand(BaseCommand):
    """
    Generic command class useful for iterating any existing command
    over all schemata. The actual command name is expected in the
    class variable COMMAND_NAME of the subclass.
    """

    def __new__(cls, *args, **kwargs):
        """
        Sets option_list and help dynamically.
        """
        obj = super(BaseTenantCommand, cls).__new__(cls, *args, **kwargs)

        app_name = get_commands()[obj.COMMAND_NAME]
        if isinstance(app_name, BaseCommand):
            # If the command is already loaded, use it directly.
            obj._original_command = app_name
        else:
            obj._original_command = load_command_class(
                app_name, obj.COMMAND_NAME
            )

        # prepend the command's original help with the info about schemata
        # iteration
        cmd = obj.COMMAND_NAME
        help = getattr(obj._original_command, 'help', 'none')
        obj.help = (
            f'Calls {cmd} for all registered schemata. You can use regular '
            f'{cmd} options.\n\nOriginal help for {cmd}:\n\n{help}'
        )

        return obj

    def add_arguments(self, parser):
        super(BaseTenantCommand, self).add_arguments(parser)
        parser.add_argument("-s", "--schema", dest="schema_name")
        parser.add_argument(
            "-p",
            "--skip-public",
            dest="skip_public",
            action="store_true",
            default=False
        )
        # use the privately held reference to the underlying command to invoke
        # the add_arguments path on this parser instance
        self._original_command.add_arguments(parser)

    def execute_command(self, tenant, command_name, *args, **options):
        verbosity = int(options.get('verbosity'))

        if verbosity >= 1:
            print()
            print(
                self.style.NOTICE("=== Switching to schema '")
                + self.style.SQL_TABLE(tenant.schema_name)
                + self.style.NOTICE("' then calling %s:" % command_name)
            )

        connection.set_tenant(tenant)

        # call the original command with the args it knows
        call_command(command_name, *args, **options)

    def handle(self, *args, **options):
        """Iterates a command over all registered schemata."""

        if options['schema_name']:
            # only run on a particular schema
            connection.set_schema_to_public()
            tenant = get_tenant_model().objects.get(
                schema_name=options['schema_name']
            )
            self.execute_command(tenant, self.COMMAND_NAME, *args, **options)
        else:
            for tenant in get_tenant_model().objects.all():
                if (
                    not options['skip_public']
                    or tenant.schema_name != get_public_schema_name()
                ):
                    self.execute_command(
                        tenant, self.COMMAND_NAME, *args, **options
                    )


class InteractiveTenantOption(object):

    def add_arguments(self, parser):
        parser.add_argument(
            "-s", "--schema", dest="schema_name", help="specify tenant schema"
        )

    def get_tenant_from_options_or_interactive(self, **options):
        TenantModel = get_tenant_model()
        tenants = TenantModel.objects.all()

        if not tenants:
            raise CommandError('There are no tenants in the system.')

        if options.get('schema_name'):
            tenant_schema = options['schema_name']
        else:
            while True:
                tenant_schema = input(
                    "Enter Tenant Schema ('?' to list schemas): "
                )
                if tenant_schema != '?':
                    break

                print('\n'.join([
                    "%s - %s" % (t.schema_name, t.domain_url,) for t in tenants
                ]))

        if tenant_schema not in [t.schema_name for t in tenants]:
            raise CommandError(
                f'Invalid tenant schema {tenant_schema}.'
            )

        return TenantModel.objects.get(schema_name=tenant_schema)


class TenantWrappedCommand(InteractiveTenantOption, BaseCommand):
    """
    Generic command class useful for running any existing command
    on a particular tenant. The actual command name is expected in the
    class variable COMMAND_NAME of the subclass.
    """

    def __new__(cls, *args, **kwargs):
        obj = super(TenantWrappedCommand, cls).__new__(cls, *args, **kwargs)
        obj.command_instance = obj.COMMAND()
        return obj

    def add_arguments(self, parser):
        super(TenantWrappedCommand, self).add_arguments(parser)
        self.command_instance.add_arguments(parser)

    def handle(self, *args, **options):
        tenant = self.get_tenant_from_options_or_interactive(**options)
        connection.set_tenant(tenant)

        self.command_instance.execute(*args, **options)


class SyncCommon(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            '--tenant',
            action='store_true',
            dest='tenant',
            default=False,
            help='Tells Django to populate only tenant applications.'
        )
        parser.add_argument(
            '--shared',
            action='store_true',
            dest='shared',
            default=False,
            help='Tells Django to populate only shared applications.'
        )
        parser.add_argument(
            '--app_label',
            action='store',
            dest='app_label',
            nargs='?',
            help='App label of an application to synchronize the state.'
        )
        parser.add_argument(
            '--migration_name',
            action='store',
            dest='migration_name',
            nargs='?',
            help=(
                'Database state will be brought to the state after that '
                'migration. Use the name "zero" to unapply all migrations.'
            )
        )
        parser.add_argument("-s", "--schema", dest="schema_name")
        parser.add_argument(
            '--executor',
            action='store',
            dest='executor',
            default=None,
            help='Executor for running migrations [standard (default)|parallel]'
        )

    def handle(self, *args, **options):
        self.sync_tenant = options.get('tenant')
        self.sync_public = options.get('shared')
        self.schema_name = options.get('schema_name')
        self.executor = options.get('executor')
        self.installed_apps = settings.INSTALLED_APPS
        self.args = args
        self.options = options

        if self.schema_name:
            if self.sync_public:
                raise CommandError(
                    'Schema should only be used with the --tenant switch.'
                )
            elif self.schema_name == get_public_schema_name():
                self.sync_public = True
            else:
                self.sync_tenant = True
        elif not self.sync_public and not self.sync_tenant:
            # no options set, sync both
            self.sync_tenant = True
            self.sync_public = True

    def _notice(self, output):
        if int(self.options.get('verbosity', 1)) >= 1:
            self.stdout.write(self.style.NOTICE(output))



from django.conf import settings
from django.db import connection
from django.http import Http404

from multi_tenant.utils import get_public_schema_name, get_tenant_model


class BaseTenantMiddleware:
    """
    Subclass and override this to achieve desired behaviour. Given a
    request, return the tenant to use. Tenant should be an instance
    of TENANT_MODEL. We have two parameters for backwards compatibility
    (the request would be enough).
    """

    TENANT_NOT_FOUND_EXCEPTION = Http404

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        TenantModel = get_tenant_model()

        try:
            # get_tenant must be implemented by extending this class.
            tenant = self.get_tenant(TenantModel, request)
        except TenantModel.DoesNotExist:
            raise self.TENANT_NOT_FOUND_EXCEPTION('No tenant found')

        if not isinstance(tenant, TenantModel):
            raise self.TENANT_NOT_FOUND_EXCEPTION(
                'Invalid tenant {!r}'.format(request.tenant)
            )

        request.tenant = tenant
        connection.set_tenant(request.tenant)

        # Do we have a public-specific urlconf?
        if (
            hasattr(settings, "PUBLIC_SCHEMA_URLCONF")
            and request.tenant.schema_name == get_public_schema_name()
        ):
            request.urlconf = settings.PUBLIC_SCHEMA_URLCONF

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response

    def get_tenant(self, model, request):
        raise NotImplementedError



from django.test import Client


class TenantClient(Client):
    def __init__(self, tenant, enforce_csrf_checks=False, **defaults):
        super().__init__(enforce_csrf_checks, **defaults)
        self.tenant = tenant

    def get_tenant_header(self, **extra):
        raise NotImplementedError(
            'Subclasess must define tenant header method.'
        )

    def get(self, *args, **kwargs):
        """Construct a GET request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().get(*args, **kwargs)

    def trace(self, *args, **kwargs):
        """Construct a TRACE request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().trace(*args, **kwargs)

    def head(self, *args, **kwargs):
        """Construct a HEAD request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().head(*args, **kwargs)

    def options(self, *args, **kwargs):
        """Construct a OPTIONS request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().options(*args, **kwargs)

    def post(self, *args, **kwargs):
        """Construct a POST request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().post(*args, **kwargs)

    def patch(self, *args, **kwargs):
        """Construct a PATCH request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().patch(*args, **kwargs)

    def put(self, *args, **kwargs):
        """Construct a PUT request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().put(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """Construct a DELETE request."""
        kwargs = self.get_tenant_header(**kwargs)
        return super().delete(*args, **kwargs)



from django.contrib.staticfiles.storage import StaticFilesStorage
from django.core.exceptions import SuspiciousOperation
from django.core.files.storage import FileSystemStorage
from django.db import connection
from django.utils._os import safe_join

__all__ = (
    'TenantStorageMixin',
    'TenantFileSystemStorage',
    'TenantStaticFilesStorage',
)


class TenantStorageMixin(object):
    """
    Mixin that can be combined with other Storage backends to colocate media
    for all tenants in distinct subdirectories.
    Using rewriting rules at the reverse proxy we can determine which content
    gets served up, while any code interactions will account for the multiple
    tenancy of the project.
    """
    def path(self, name):
        """
        Look for files in subdirectory of MEDIA_ROOT using the tenant's
        codename value as the specifier.
        """
        if name is None:
            name = ''
        try:
            location = safe_join(self.location, connection.tenant.codename)
        except AttributeError:
            location = self.location
        try:
            path = safe_join(location, name)
        except ValueError:
            raise SuspiciousOperation(f'Attempted access to {name!r} denied.')
        return path


class TenantFileSystemStorage(TenantStorageMixin, FileSystemStorage):
    """
    Implementation that extends core Django's FileSystemStorage.
    """


class TenantStaticFilesStorage(TenantStorageMixin, StaticFilesStorage):
    """
    Implementation that extends core Django's StaticFilesStorage.
    """



import django.db.utils
import psycopg2
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from tenant_schemas.postgresql_backend.base import (
    DatabaseWrapper as TenantDatabaseWrapper
)
from tenant_schemas.postgresql_backend.base import (
    FakeTenant,
    _check_schema_name
)
from tenant_schemas.utils import get_public_schema_name, get_limit_set_calls

ORIGINAL_BACKEND = getattr(
    settings, 'ORIGINAL_BACKEND', 'django.db.backends.postgresql_psycopg2'
)
# Django 1.9+ takes care to rename the default
# backend to 'django.db.backends.postgresql'
original_backend = django.db.utils.load_backend(ORIGINAL_BACKEND)

EXTRA_SEARCH_PATHS = getattr(settings, 'PG_EXTRA_SEARCH_PATHS', [])


class DatabaseWrapper(TenantDatabaseWrapper):

    def set_tenant(self, tenant, include_public=True):
        """
        Main API method to current database schema,
        but it does not actually modify the db connection.
        """
        self.main_tenant = (
            None if isinstance(self.tenant, FakeTenant) else tenant.tenant
        )
        return super().set_tenant(tenant, include_public=include_public)

    def _cursor(self, name=None):
        """
        Here it happens. We hope every Django db operation using PostgreSQL
        must go through this to get the cursor handle. We change the path.
        """
        cursor = super(original_backend.DatabaseWrapper, self)._cursor()

        # optionally limit the number of executions - under load, the execution
        # of `set search_path` can be quite time consuming
        if (not get_limit_set_calls()) or not self.search_path_set:
            # Actual search_path modification for the cursor. Database will
            # search schemata from left to right when looking for the object
            # (table, index, sequence, etc.).
            if not self.schema_name:
                raise ImproperlyConfigured(
                    'Database schema not set. Did you forget '
                    'to call set_schema() or set_tenant()?'
                )

            search_paths = self._get_search_path()
            if name:
                # Named cursor can only be used once
                cursor_for_search_path = self.connection.cursor()
            else:
                # Reuse
                cursor_for_search_path = cursor

            # In the event that an error already happened in this
            # transaction and we are going to rollback we should just ignore
            # database error when setting the search_path if the next
            # instruction is not a rollback it will just fail also, so
            # we do not have to worry that it's not the good one
            try:
                cursor_for_search_path.execute(
                    'SET search_path = {0}'.format(', '.join(search_paths))
                )
            except (django.db.utils.DatabaseError, psycopg2.InternalError):
                self.search_path_set = False
            else:
                self.search_path_set = True

            if name:
                cursor_for_search_path.close()

        return cursor

    def _get_search_path(self):
        """Build the search path according to the current schema."""

        _check_schema_name(self.schema_name)
        public_schema_name = get_public_schema_name()
        search_paths = [self.schema_name]

        if self.schema_name != public_schema_name:
            if self.main_tenant:
                search_paths.append(self.main_tenant.schema_name)

            if self.include_public_schema:
                search_paths.append(public_schema_name)

        search_paths.extend(EXTRA_SEARCH_PATHS)
        return search_paths



import sys

from django.core.management.commands.migrate import Command as MigrateCommand
from django.db import transaction
from tenant_schemas.utils import get_public_schema_name, get_tenant_model


def run_migrations(
    args, options, executor_codename, tenant, allow_atomic=True
):
    from django.core.management import color
    from django.core.management.base import OutputWrapper
    from django.db import connection

    style = color.color_style()
    TenantModel = get_tenant_model()
    schema_name = tenant.schema_name \
        if isinstance(tenant, TenantModel) else tenant

    def style_func(msg):
        return '[%s:%s] %s' % (
            style.NOTICE(executor_codename),
            style.NOTICE(schema_name),
            msg
        )

    stdout = OutputWrapper(sys.stdout)
    stdout.style_func = style_func
    stderr = OutputWrapper(sys.stderr)
    stderr.style_func = style_func
    if int(options.get('verbosity', 1)) >= 1:
        stdout.write(style.NOTICE(
            f'=== Running migrate for schema {schema_name}'
        ))

    if isinstance(tenant, TenantModel):
        connection.set_tenant(tenant)
    elif isinstance(tenant, str):
        connection.set_schema(tenant)

    MigrateCommand(stdout=stdout, stderr=stderr).execute(*args, **options)

    try:
        transaction.commit()
        connection.close()
        connection.connection = None
    except transaction.TransactionManagementError:
        if not allow_atomic:
            raise

        # We are in atomic transaction, don't close connections
        pass

    connection.set_schema_to_public()


class MigrationExecutor(object):
    codename = None

    def __init__(self, args, options):
        self.args = args
        self.options = options

    def run_migrations(self, tenants):
        public_schema_name = get_public_schema_name()

        for idx, tenant in enumerate(tenants):
            schema_name = tenant.schema_name \
                if isinstance(tenant, get_tenant_model()) else tenant
            if schema_name == public_schema_name:
                run_migrations(self.args, self.options, self.codename, tenant)
                tenants.pop(idx)
                break

        self.run_tenant_migrations(tenants)

    def run_tenant_migrations(self, tenants):
        raise NotImplementedError



#!/usr/bin/env python

from setuptools import find_packages, setup

VERSION = '0.1.3.1'

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='django-multi-tenants',
    version=VERSION,
    author='Cristian Restrepo',
    author_email='crstnrrm@gmail.com',
    packages=find_packages(),
    scripts=[],
    url='https://gitlab.com/crstnrm/django-multi-tenants',
    license='MIT',
    description='Multi tenant support for Django using PostgreSQL schemas.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
    ],
    install_requires=[
        'Django>=2.2.0',
        'django-tenant-schemas'
    ],
    python_requires='>=3.6',
    zip_safe=False,
)



from multi_tenant.migration_executors.base import (
    MigrationExecutor,
    run_migrations
)


class StandardExecutor(MigrationExecutor):
    codename = 'standard'

    def run_tenant_migrations(self, tenants):
        for tenant in tenants:
            run_migrations(self.args, self.options, self.codename, tenant)
